#include <stdio.h>
#include <varargs.h>
#include "gba.h"

volatile s16 tx_len,rx_len;
static u32 tx_buf[3+256/4];
static u32 rx_buf[256/4];
static u32 idx;

void pio_init(){
	idx=0;
	*(vu16 *)REG_RCNT  = 0;
	*(vu16 *)REG_SIOCNT = SIO_32BIT_MODE;
	*(vu16 *)REG_SIOCNT |= SIO_IF_ENABLE
						|  SIO_START;
	*((vu16 *)REG_SIODATA32+0)=0xaaaa;
	*((vu16 *)REG_SIODATA32+1)=0x5555;
}

static void pio_wait(){
	for(;;){
		if((tx_len==-2)&&(rx_len==0)) break;
	}
}


s32 PcOpen(s8 *filename,s32 mode){
	u32 i;
	u32 len;
	tx_buf[0]=0xf0f0f001;
	for(len=0;;len++) if(filename[len]==0) break;
	len=(strlen(filename)+1+3)/4;
	tx_buf[1]=len;
	for(i=0;i<len;i++){
		tx_buf[2+i]=*((u32 *)(filename+i*4));
	}
	tx_buf[2+len]=mode;
	rx_len=1;
	tx_len=3+len;
	pio_wait();
	return((s32)rx_buf[0]);
}

s32 PcClose(s32 fh){
	tx_buf[0]=0xf0f0f002;
	tx_buf[1]=(u32)fh;
	rx_len=0;
	tx_len=2;
	pio_wait();
	return(0);
}

s32 PcRead(s32 fh,void *buffer,s32 size){
	s32 len;
	s32 zan,ofs;
	tx_buf[0]=0xf0f0f003;
	tx_buf[1]=(u32)fh;
	tx_buf[2]=(u32)size;
	rx_len=1;
	tx_len=3;
	pio_wait();
	size=(s32)rx_buf[0];
	if(size>0){
		zan=size;
		ofs=0;
		for(;;){
			if(zan>256){
				len=256;
			}
			else{
				len=zan;
			}
			rx_len=(len+3)/4;
			tx_len=-1;
			pio_wait();
			DmaCopy(3,rx_buf,(u32)buffer+ofs,(len+3)/4*4,16);
			ofs+=len;
			zan-=len;
			if(zan==0) break;
		}
	}
	return(size);
}

s32 PcWrite(s32 fh,void *buffer,s32 size){
	s32 len,zan,ofs;
	zan=size;
	ofs=0;
	for(;;){
		if(zan>256){
			len=256;
		}
		else{
			len=zan;
		}
		tx_buf[0]=0xf0f0f004;
		tx_buf[1]=(u32)fh;
		tx_buf[2]=(u32)len;
		DmaCopy(3,(u32)buffer+ofs,tx_buf+3,(len+3)/4*4,16);
		rx_len=0;
		tx_len=3+(len+3)/4;
		pio_wait();
		ofs+=len;
		zan-=len;
		if(zan==0) break;
	}
	return(size);
}

s32 PcLseek(s32 fh,s32 offset,s32 where){
	tx_buf[0]=0xf0f0f005;
	tx_buf[1]=(u32)fh;
	tx_buf[2]=(u32)offset;
	tx_buf[3]=(u32)where;
	rx_len=1;
	tx_len=4;
	pio_wait();
	return((s32)rx_buf[0]);
}

static s32 _udeci(u32 u,s8 *s,s32 i){
	u32 div,flg;
	div=1000000000;
	flg=0;
	for(;;){
		if(u/div){
			s[i++]='0'+(u/div);
			flg=1;
			u-=(u/div)*div;
		}
		else if(flg || (div==1)){
			s[i++]='0';
		}
		div=div/10;
		if(div==0) break;
	}
	return(i);
}

const static s8 hexchar[16]={
'0','1','2','3','4','5','6','7',
'8','9','A','B','C','D','E','F'
};

static s32 _hex(u32 u,s8 *s,s32 i){
	u32 div,flg;
	div=0x10000000;
	flg=0;
	for(;;){
		if(u/div){
			s[i++]=hexchar[u/div];
			flg=1;
			u-=(u/div)*div;
		}
		else if(flg || (div==1)){
			s[i++]='0';
		}
		div=div/16;
		if(div==0) break;
	}
	return(i);
}

//	%d:POi
//	%u:ȂPOi
//	%x:PUi
//	%s:
//	%c:
//	\n:s

s32 PcPrintf(va_alist) va_dcl{
	va_list args;
	s32 i,d;
	u32 u;
	s8 *ss;
	s8 *fmt;
	s8 *s;

	tx_buf[0]=0xf0f0f000;
	s=(u8 *)(tx_buf+2);
	va_start(args);
	fmt=va_arg(args,char *);
	for(i=0;;){
		if(*fmt=='%'){
			switch(fmt[1]){
			case 'd':		//	%d:POi
				fmt+=2;
				d=va_arg(args,s32);
				if(d<0){
					s[i++]='-';
					u=-d;
				}
				else{
					u=d;
				}
				i=_udeci(u,s,i);
				break;
			case 'u':		//	%u:ȂPOi
				fmt+=2;
				u=va_arg(args,u32);
				i=_udeci(u,s,i);
				break;
			case 'x':		//	%x:PUi
				fmt+=2;
				u=va_arg(args,u32);
				i=_hex(u,s,i);
				break;
			case 's':		//	%s:
				fmt+=2;
				ss=va_arg(args,s8 *);
				for(d=0;;d++){
					if(ss[d]==0) break;
					s[i++]=ss[d];
				}
				break;
			case 'c':		//	%c:
				fmt+=2;
				d=va_arg(args,s32);
				s[i++]=d;
				break;
			default:
				s[i++]=*fmt;
				fmt++;
				break;
			}
		}
		else{
			if((i>240)||(*fmt==0)){
				s[i++]=0;
				break;
			}
			s[i++]=*fmt;
			fmt++;
		}
	}
	va_end(args);
	tx_buf[1]=(i+3)/4;
	rx_len=0;
	tx_len=2+(i+3)/4;
	pio_wait();
	return(0);
}

void SioIntr(void){
	u32 d;
	d=*((vu16 *)REG_SIODATA32+0);
	d+=*((vu16 *)REG_SIODATA32+1)<<16;
	*((vu16 *)REG_SIODATA32+0)=0xaaaa;
	*((vu16 *)REG_SIODATA32+1)=0x5555;
	if((tx_len==-1)&&(rx_len!=0)){
		rx_buf[idx++]=d;
		if(--rx_len==0){
			idx=0;
			tx_len=-2;
		}
	}
	if(tx_len>=0){//faƂ`uq̏𓯊邽߂PoCg_~[𑗂
		if(tx_len--==0){
			idx=0;
			if(rx_len==0) tx_len=-2;
		}
		else{
			*((vu16 *)REG_SIODATA32+0)=tx_buf[idx];
			*((vu16 *)REG_SIODATA32+1)=tx_buf[idx]>>16;
			++idx;
		}
	}
    *(vu16 *)REG_SIOCNT |= SIO_START;        // ʐMX^[g
}
